/*
 * Copyright AMD,NAG 2003-2017
 */

/*
  This is the RNG header file. It contains function prototypes
  to allow a C programmer to call RNG routines via their C
  or Fortran interfaces.

  C interfaces to RNG routines differ from FORTRAN interfaces in
  the following major respects:

#if defined(WIN32_VF)
    (i) The FORTRAN interface names are in upper case.
#elif defined(WIN_IFORT) // ! WIN32
    (i) The FORTRAN interface names are in upper case.
#else // ! WIN32
    (i) The FORTRAN interface names are appended by an underscore.
#endif // ! WIN32
   (ii) The C interfaces contain no workspace arguments; all
        workspace memory is allocated internally.

  (iii) Scalar input arguments are passed by value in C interfaces.
        FORTRAN interfaces pass all arguments (except for character
        string "length" arguments that are normally hidden from
        FORTRAN programmers) by reference.

   (iv) Most arguments that are passed as character string pointers
        to FORTRAN interfaces are passed by value as single
        characters to C interfaces. The character string "length"
        arguments of FORTRAN interfaces are not required in the
        C interfaces.

  It is important to note that in both the FORTRAN and C interfaces,
  2-dimensional arrays are assumed to be stored in column-major order.
  e.g. the matrix A = [ 1.0 2.0 ]
                      [ 3.0 4.0 ]
  would be stored in memory as 1.0, 3.0, 2.0, 4.0. This storage order
  corresponds to a FORTRAN-style 2-D array declaration A(2,2), but not
  to an array declared as a[2][2] in C which would be stored in
  row-major order as 1.0, 2.0, 3.0, 4.0.

  As an example, compare the FORTRAN and C interfaces of LAPACK
  routine dsytrf as implemented in RNG.

#if defined(WIN32_VF)

  FORTRAN:
    void DSYTRF(char *uplo, rng_strlen_t uplo_len, rng_int_t *n, double *a, rng_int_t *lda,
                rng_int_t *ipiv, double *work, rng_int_t *lwork, rng_int_t *info);
#elif defined(WIN_IFORT) // ! WIN32

  FORTRAN:
    void DSYTRF(char *uplo, rng_int_t *n, double *a, rng_int_t *lda, rng_int_t *ipiv,
                double *work, rng_int_t *lwork, rng_int_t *info, rng_strlen_t uplo_len);
#else // ! WIN32

  FORTRAN:
    void dsytrf_(char *uplo, rng_int_t *n, double *a, rng_int_t *lda, rng_int_t *ipiv,
                 double *work, rng_int_t *lwork, rng_int_t *info, rng_strlen_t uplo_len);
#endif // ! WIN32

  C:
    void dsytrf(char uplo, rng_int_t n, double *a, rng_int_t lda, rng_int_t *ipiv, rng_int_t *info);

  C code calling both the above variants might look like this:

    double *a;
    rng_int_t *ipiv;
    double *work;
    rng_int_t n, lda, lwork, info;

    //
    // Assume that all arrays and variables are allocated and initialized
    // as required by dsytrf.
#if defined(WIN32_VF)
    //
    // Call the FORTRAN version of dsytrf. The first argument is a character
    // string, and the second argument is the length of that string. The input
    // scalar arguments n, lda and lwork, as well as the output scalar
    // argument info, are all passed by reference.
    DSYTRF("Upper", 5, &n, a, &lda, ipiv, work, &lwork, &info);
#elif defined(WIN_IFORT) // ! WIN32
    //
    // Call the FORTRAN version of dsytrf. The first argument is a character
    // string, and the last argument is the length of that string. The input
    // scalar arguments n, lda and lwork, as well as the output scalar
    // argument info, are all passed by reference.
    DSYTRF("Upper", &n, a, &lda, ipiv, work, &lwork, &info, 5);
#else // ! WIN32
    //
    // Call the FORTRAN version of dsytrf. The first argument is a character
    // string, and the last argument is the length of that string. The input
    // scalar arguments n, lda and lwork, as well as the output scalar
    // argument info, are all passed by reference.
    dsytrf_("Upper", &n, a, &lda, ipiv, work, &lwork, &info, 5);
#endif // ! WIN32
    //
    // Call the C version of dsytrf. The first argument is a character,
    // workspace is not required, and input scalar arguments n and lda are
    // passed by value. Output scalar argument info is passed by reference.
    dsytrf('U', n, a, lda, ipiv, &info);

*/

#ifndef _RNG_H
#define _RNG_H

/* Under Windows math.h defines "complex" to mean "_complex". */
#include <math.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* ------------------------------------------------------------------ */

/* All lines containing _RNG_INT_T or INTEGER64 get          _RNG_INT_T
   filtered out of the distributed copy of rng.h             _RNG_INT_T */
/* _RNG_INT_T The RNG integer type, either 32-bit or 64-bit */
#ifndef _RNG_INT_T
#define _RNG_INT_T
#ifdef INTEGER64
/* _RNG_INT_T  N.B. even in 64-bit integer variants of RNG, hidden
   _RNG_INT_T  Fortran string length arguments (type rng_strlen_t)
   _RNG_INT_T  are of type int rather than long */
#ifdef WIN64 /* INTEGER64 */
  typedef __int64 rng_int_t;             /* _RNG_INT_T */
  typedef unsigned __int64 rng_uint_t;   /* _RNG_INT_T */
  typedef int rng_strlen_t;              /* _RNG_INT_T */
#else /* INTEGER64 */
  typedef long rng_int_t;                /* _RNG_INT_T */
  typedef unsigned long rng_uint_t;      /* _RNG_INT_T */
  typedef int rng_strlen_t;              /* _RNG_INT_T */
#endif /* INTEGER64 */
#else /* ! INTEGER64 */
typedef int rng_int_t;                 /* _RNG_INT_T */
typedef unsigned int rng_uint_t;       /* _RNG_INT_T */
typedef int rng_strlen_t;              /* _RNG_INT_T */
#endif /* ! INTEGER64 */
#endif /* _RNG_INT_T */


#if defined(WIN32_VF)
  #ifdef USE_CDECL
    #define RNGCDECL __cdecl
    #define RNGSTDCALL __cdecl
  #else /* USE_CDECL */
    #define RNGCDECL _stdcall
    #define RNGSTDCALL _stdcall
  #endif /* USE_CDECL */
#elif defined(WIN32_PVF) /* ! WIN32_VF */
  #define RNGCDECL
  #define RNGSTDCALL
#else /* ! WIN32_VF */
  #define RNGSTDCALL
  #define RNGCDECL
#endif /* ! WIN32_VF */


typedef void (RNGSTDCALL * RNG_DRANDINITIALIZEUSER_UINI)(rng_int_t *, rng_int_t *, rng_int_t *, rng_int_t *, rng_int_t *, rng_int_t *, rng_int_t *);
typedef void (RNGSTDCALL * RNG_DRANDINITIALIZEUSER_UGEN)(rng_int_t *, rng_int_t *, double *, rng_int_t *);
typedef void (RNGSTDCALL * RNG_SRANDINITIALIZEUSER_UINI)(rng_int_t *, rng_int_t *, rng_int_t *, rng_int_t *, rng_int_t *, rng_int_t *, rng_int_t *);
typedef void (RNGSTDCALL * RNG_SRANDINITIALIZEUSER_UGEN)(rng_int_t *, rng_int_t *, float *, rng_int_t *);


/* Below are prototypes for the C interfaces to RNG routines */

/* Random number generators */
void RNGCDECL drandbeta(rng_int_t n, double a, double b, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL drandbinomial(rng_int_t n, rng_int_t m, double p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL drandbinomialreference(rng_int_t m, double p, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL drandblumblumshub(rng_int_t n, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL drandcauchy(rng_int_t n, double a, double b, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL drandchisquared(rng_int_t n, rng_int_t df, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL dranddiscreteuniform(rng_int_t n, rng_int_t a, rng_int_t b, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL drandexponential(rng_int_t n, double a, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL drandf(rng_int_t n, rng_int_t df1, rng_int_t df2, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL drandgamma(rng_int_t n, double a, double b, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL drandgaussian(rng_int_t n, double xmu, double var, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL drandgeneraldiscrete(rng_int_t n, double *ref, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL drandgeometric(rng_int_t n, double p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL drandgeometricreference(double p, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL drandhypergeometric(rng_int_t n, rng_int_t np, rng_int_t ns, rng_int_t m, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL drandhypergeometricreference(rng_int_t np, rng_int_t ns, rng_int_t m, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL drandinitialize(rng_int_t genid, rng_int_t subid, rng_int_t *seed, rng_int_t *lseed, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void RNGCDECL drandinitialize_secrng(rng_int_t genid, rng_int_t subid, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void RNGCDECL drandinitializebbs(rng_int_t nbits, rng_int_t lp, rng_int_t *p, rng_int_t lq, rng_int_t *q, rng_int_t ls, rng_int_t *s, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void RNGCDECL drandinitializeuser(RNG_DRANDINITIALIZEUSER_UINI uini, RNG_DRANDINITIALIZEUSER_UGEN ugen, rng_int_t genid, rng_int_t subid, rng_int_t *seed, rng_int_t *lseed, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void RNGCDECL drandleapfrog(rng_int_t n, rng_int_t k, rng_int_t *state, rng_int_t *info);
void RNGCDECL drandlogistic(rng_int_t n, double a, double b, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL drandlognormal(rng_int_t n, double xmu, double var, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL drandmultinomial(rng_int_t n, rng_int_t m, double *p, rng_int_t k, rng_int_t *state, rng_int_t *x, rng_int_t ldx, rng_int_t *info);
void RNGCDECL drandmultinormalreference(rng_int_t m, double *xmu, double *c, rng_int_t ldc, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL drandmultinormalr(rng_int_t n, double *ref, rng_int_t *state, double *x, rng_int_t ldx, rng_int_t *info);
void RNGCDECL drandmultistudentstreference(rng_int_t m, rng_int_t df, double *xmu, double *c, rng_int_t ldc, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL drandmultistudentstr(rng_int_t n, double *ref, rng_int_t *state, double *x, rng_int_t ldx, rng_int_t *info);
void RNGCDECL drandnegativebinomial(rng_int_t n, rng_int_t m, double p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL drandnegativebinomialreference(rng_int_t m, double p, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL drandpoisson(rng_int_t n, double lambda, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL drandpoissonreference(double lambda, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL drandskipahead(rng_int_t n, rng_int_t *state, rng_int_t *info);
void RNGCDECL drandstudentst(rng_int_t n, rng_int_t df, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL drandtriangular(rng_int_t n, double xmin, double xmed, double xmax, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL dranduniform(rng_int_t n, double a, double b, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL drandvonmises(rng_int_t n, double vk, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL drandweibull(rng_int_t n, double a, double b, rng_int_t *state, double *x, rng_int_t *info);
void RNGCDECL drandmultinormal(rng_int_t n, rng_int_t m, double *xmu, double *c, rng_int_t ldc, rng_int_t *state, double *x, rng_int_t ldx, rng_int_t *info);
void RNGCDECL drandmultistudentst(rng_int_t n, rng_int_t m, rng_int_t df, double *xmu, double *c, rng_int_t ldc, rng_int_t *state, double *x, rng_int_t ldx, rng_int_t *info);
void RNGCDECL srandbeta(rng_int_t n, float a, float b, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL srandblumblumshub(rng_int_t n, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL srandbinomial(rng_int_t n, rng_int_t m, float p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL srandbinomialreference(rng_int_t m, float p, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL srandcauchy(rng_int_t n, float a, float b, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL srandchisquared(rng_int_t n, rng_int_t df, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL sranddiscreteuniform(rng_int_t n, rng_int_t a, rng_int_t b, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL srandexponential(rng_int_t n, float a, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL srandf(rng_int_t n, rng_int_t df1, rng_int_t df2, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL srandgamma(rng_int_t n, float a, float b, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL srandgaussian(rng_int_t n, float xmu, float var, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL srandgeneraldiscrete(rng_int_t n, float *ref, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL srandgeometric(rng_int_t n, float p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL srandgeometricreference(float p, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL srandhypergeometric(rng_int_t n, rng_int_t np, rng_int_t ns, rng_int_t m, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL srandhypergeometricreference(rng_int_t np, rng_int_t ns, rng_int_t m, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL srandinitialize(rng_int_t genid, rng_int_t subid, rng_int_t *seed, rng_int_t *lseed, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void RNGCDECL srandinitializebbs(rng_int_t nbits, rng_int_t lp, rng_int_t *p, rng_int_t lq, rng_int_t *q, rng_int_t ls, rng_int_t *s, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void RNGCDECL srandinitializeuser(RNG_SRANDINITIALIZEUSER_UINI uini, RNG_SRANDINITIALIZEUSER_UGEN ugen, rng_int_t genid, rng_int_t subid, rng_int_t *seed, rng_int_t *lseed, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void RNGCDECL srandleapfrog(rng_int_t n, rng_int_t k, rng_int_t *state, rng_int_t *info);
void RNGCDECL srandlogistic(rng_int_t n, float a, float b, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL srandlognormal(rng_int_t n, float xmu, float var, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL srandmultinomial(rng_int_t n, rng_int_t m, float *p, rng_int_t k, rng_int_t *state, rng_int_t *x, rng_int_t ldx, rng_int_t *info);
void RNGCDECL srandmultinormalreference(rng_int_t m, float *xmu, float *c, rng_int_t ldc, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL srandmultinormalr(rng_int_t n, float *ref, rng_int_t *state, float *x, rng_int_t ldx, rng_int_t *info);
void RNGCDECL srandmultistudentstreference(rng_int_t m, rng_int_t df, float *xmu, float *c, rng_int_t ldc, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL srandmultistudentstr(rng_int_t n, float *ref, rng_int_t *state, float *x, rng_int_t ldx, rng_int_t *info);
void RNGCDECL srandnegativebinomial(rng_int_t n, rng_int_t m, float p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL srandnegativebinomialreference(rng_int_t m, float p, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL srandpoisson(rng_int_t n, float lambda, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGCDECL srandpoissonreference(float lambda, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGCDECL srandskipahead(rng_int_t n, rng_int_t *state, rng_int_t *info);
void RNGCDECL srandstudentst(rng_int_t n, rng_int_t df, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL srandtriangular(rng_int_t n, float xmin, float xmed, float xmax, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL sranduniform(rng_int_t n, float a, float b, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL srandvonmises(rng_int_t n, float vk, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL srandweibull(rng_int_t n, float a, float b, rng_int_t *state, float *x, rng_int_t *info);
void RNGCDECL srandmultinormal(rng_int_t n, rng_int_t m, float *xmu, float *c, rng_int_t ldc, rng_int_t *state, float *x, rng_int_t ldx, rng_int_t *info);
void RNGCDECL srandmultistudentst(rng_int_t n, rng_int_t m, rng_int_t df, float *xmu, float *c, rng_int_t ldc, rng_int_t *state, float *x, rng_int_t ldx, rng_int_t *info);

/* RNG version information */
const char* RNGCDECL get_rngversion(void);




/* Below are prototypes for the FORTRAN interfaces(with underscore) to RNG routines */

void drandbeta_(rng_int_t *n, double *a, double *b, rng_int_t *state, double *x, rng_int_t *info);
void drandbinomial_(rng_int_t *n, rng_int_t *m, double *p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void drandbinomialreference_(rng_int_t *m, double *p, double *ref, rng_int_t *lref, rng_int_t *info);
void drandblumblumshub_(rng_int_t *n, rng_int_t *state, double *x, rng_int_t *info);
void drandcauchy_(rng_int_t *n, double *a, double *b, rng_int_t *state, double *x, rng_int_t *info);
void drandchisquared_(rng_int_t *n, rng_int_t *df, rng_int_t *state, double *x, rng_int_t *info);
void dranddiscreteuniform_(rng_int_t *n, rng_int_t *a, rng_int_t *b, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void drandexponential_(rng_int_t *n, double *a, rng_int_t *state, double *x, rng_int_t *info);
void drandf_(rng_int_t *n, rng_int_t *df1, rng_int_t *df2, rng_int_t *state, double *x, rng_int_t *info);
void drandgamma_(rng_int_t *n, double *a, double *b, rng_int_t *state, double *x, rng_int_t *info);
void drandgaussian_(rng_int_t *n, double *xmu, double *var, rng_int_t *state, double *x, rng_int_t *info);
void drandgeneraldiscrete_(rng_int_t *n, double *ref, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void drandgeometric_(rng_int_t *n, double *p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void drandgeometricreference_(double *p, double *ref, rng_int_t *lref, rng_int_t *info);
void drandhypergeometric_(rng_int_t *n, rng_int_t *np, rng_int_t *ns, rng_int_t *m, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void drandhypergeometricreference_(rng_int_t *np, rng_int_t *ns, rng_int_t *m, double *ref, rng_int_t *lref, rng_int_t *info);
void drandinitialize_(rng_int_t *genid, rng_int_t *subid, rng_int_t *seed, rng_int_t *lseed, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void drandinitializebbs_(rng_int_t *nbits, rng_int_t *lp, rng_int_t *p, rng_int_t *lq, rng_int_t *q, rng_int_t *ls, rng_int_t *s, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void drandinitializeuser_(RNG_DRANDINITIALIZEUSER_UINI uini, RNG_DRANDINITIALIZEUSER_UGEN ugen, rng_int_t genid, rng_int_t subid, rng_int_t *seed, rng_int_t *lseed, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void drandleapfrog_(rng_int_t *n, rng_int_t *k, rng_int_t *state, rng_int_t *info);
void drandlogistic_(rng_int_t *n, double *a, double *b, rng_int_t *state, double *x, rng_int_t *info);
void drandlognormal_(rng_int_t *n, double *xmu, double *var, rng_int_t *state, double *x, rng_int_t *info);
void drandmultinomial_(rng_int_t *n, rng_int_t *m, double *p, rng_int_t *k, rng_int_t *state, rng_int_t *x, rng_int_t *ldx, rng_int_t *info);
void drandmultinormalreference_(rng_int_t *m, double *xmu, double *c, rng_int_t *ldc, double *ref, rng_int_t *lref, rng_int_t *info);
void drandmultinormalr_(rng_int_t *n, double *ref, rng_int_t *state, double *x, rng_int_t *ldx, rng_int_t *info);
void drandmultistudentstreference_(rng_int_t *m, rng_int_t *df, double *xmu, double *c, rng_int_t *ldc, double *ref, rng_int_t *lref, rng_int_t *info);
void drandmultistudentstr_(rng_int_t *n, double *ref, rng_int_t *state, double *x, rng_int_t *ldx, rng_int_t *info);
void drandnegativebinomial_(rng_int_t *n, rng_int_t *m, double *p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void drandnegativebinomialreference_(rng_int_t *m, double *p, double *ref, rng_int_t *lref, rng_int_t *info);
void drandpoisson_(rng_int_t *n, double *lambda, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void drandpoissonreference_(double *lambda, double *ref, rng_int_t *lref, rng_int_t *info);
void drandskipahead_(rng_int_t *n, rng_int_t *state, rng_int_t *info);
void drandstudentst_(rng_int_t *n, rng_int_t *df, rng_int_t *state, double *x, rng_int_t *info);
void drandtriangular_(rng_int_t *n, double *xmin, double *xmed, double *xmax, rng_int_t *state, double *x, rng_int_t *info);
void dranduniform_(rng_int_t *n, double *a, double *b, rng_int_t *state, double *x, rng_int_t *info);
void drandvonmises_(rng_int_t *n, double *vk, rng_int_t *state, double *x, rng_int_t *info);
void drandweibull_(rng_int_t *n, double *a, double *b, rng_int_t *state, double *x, rng_int_t *info);
void drandmultinormal_(rng_int_t *n, rng_int_t *m, double *xmu, double *c, rng_int_t *ldc, rng_int_t *state, double *x, rng_int_t *ldx, rng_int_t *info);
void drandmultistudentst_(rng_int_t *n, rng_int_t *m, rng_int_t *df, double *xmu, double *c, rng_int_t *ldc, rng_int_t *state, double *x, rng_int_t *ldx, rng_int_t *info);
void srandbeta_(rng_int_t *n, float *a, float *b, rng_int_t *state, float *x, rng_int_t *info);
void srandblumblumshub_(rng_int_t *n, rng_int_t *state, float *x, rng_int_t *info);
void srandbinomial_(rng_int_t *n, rng_int_t *m, float *p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void srandbinomialreference_(rng_int_t *m, float *p, float *ref, rng_int_t *lref, rng_int_t *info);
void srandcauchy_(rng_int_t *n, float *a, float *b, rng_int_t *state, float *x, rng_int_t *info);
void srandchisquared_(rng_int_t *n, rng_int_t *df, rng_int_t *state, float *x, rng_int_t *info);
void sranddiscreteuniform_(rng_int_t *n, rng_int_t *a, rng_int_t *b, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void srandexponential_(rng_int_t *n, float *a, rng_int_t *state, float *x, rng_int_t *info);
void srandf_(rng_int_t *n, rng_int_t *df1, rng_int_t *df2, rng_int_t *state, float *x, rng_int_t *info);
void srandgamma_(rng_int_t *n, float *a, float *b, rng_int_t *state, float *x, rng_int_t *info);
void srandgaussian_(rng_int_t *n, float *xmu, float *var, rng_int_t *state, float *x, rng_int_t *info);
void srandgeneraldiscrete_(rng_int_t *n, float *ref, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void srandgeometric_(rng_int_t *n, float *p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void srandgeometricreference_(float *p, float *ref, rng_int_t *lref, rng_int_t *info);
void srandhypergeometric_(rng_int_t *n, rng_int_t *np, rng_int_t *ns, rng_int_t *m, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void srandhypergeometricreference_(rng_int_t *np, rng_int_t *ns, rng_int_t *m, float *ref, rng_int_t *lref, rng_int_t *info);
void srandinitialize_(rng_int_t *genid, rng_int_t *subid, rng_int_t *seed, rng_int_t *lseed, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void srandinitializebbs_(rng_int_t *nbits, rng_int_t *lp, rng_int_t *p, rng_int_t *lq, rng_int_t *q, rng_int_t *ls, rng_int_t *s, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void srandinitializeuser_(RNG_SRANDINITIALIZEUSER_UINI uini, RNG_SRANDINITIALIZEUSER_UGEN ugen, rng_int_t genid, rng_int_t subid, rng_int_t *seed, rng_int_t *lseed, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void srandleapfrog_(rng_int_t *n, rng_int_t *k, rng_int_t *state, rng_int_t *info);
void srandlogistic_(rng_int_t *n, float *a, float *b, rng_int_t *state, float *x, rng_int_t *info);
void srandlognormal_(rng_int_t *n, float *xmu, float *var, rng_int_t *state, float *x, rng_int_t *info);
void srandmultinomial_(rng_int_t *n, rng_int_t *m, float *p, rng_int_t *k, rng_int_t *state, rng_int_t *x, rng_int_t *ldx, rng_int_t *info);
void srandmultinormalreference_(rng_int_t *m, float *xmu, float *c, rng_int_t *ldc, float *ref, rng_int_t *lref, rng_int_t *info);
void srandmultinormalr_(rng_int_t *n, float *ref, rng_int_t *state, float *x, rng_int_t *ldx, rng_int_t *info);
void srandmultistudentstreference_(rng_int_t *m, rng_int_t *df, float *xmu, float *c, rng_int_t *ldc, float *ref, rng_int_t *lref, rng_int_t *info);
void srandmultistudentstr_(rng_int_t *n, float *ref, rng_int_t *state, float *x, rng_int_t *ldx, rng_int_t *info);
void srandnegativebinomial_(rng_int_t *n, rng_int_t *m, float *p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void srandnegativebinomialreference_(rng_int_t *m, float *p, float *ref, rng_int_t *lref, rng_int_t *info);
void srandpoisson_(rng_int_t *n, float *lambda, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void srandpoissonreference_(float *lambda, float *ref, rng_int_t *lref, rng_int_t *info);
void srandskipahead_(rng_int_t *n, rng_int_t *state, rng_int_t *info);
void srandstudentst_(rng_int_t *n, rng_int_t *df, rng_int_t *state, float *x, rng_int_t *info);
void srandtriangular_(rng_int_t *n, float *xmin, float *xmed, float *xmax, rng_int_t *state, float *x, rng_int_t *info);
void sranduniform_(rng_int_t *n, float *a, float *b, rng_int_t *state, float *x, rng_int_t *info);
void srandvonmises_(rng_int_t *n, float *vk, rng_int_t *state, float *x, rng_int_t *info);
void srandweibull_(rng_int_t *n, float *a, float *b, rng_int_t *state, float *x, rng_int_t *info);
void srandmultinormal_(rng_int_t *n, rng_int_t *m, float *xmu, float *c, rng_int_t *ldc, rng_int_t *state, float *x, rng_int_t *ldx, rng_int_t *info);
void srandmultistudentst_(rng_int_t *n, rng_int_t *m, rng_int_t *df, float *xmu, float *c, rng_int_t *ldc, rng_int_t *state, float *x, rng_int_t *ldx, rng_int_t *info);

/* RNG version information */
const char *get_rngversion_(void);




/* Below are prototypes for the FORTRAN interfaces(with upper case) to RNG routines */

void RNGSTDCALL DRANDBETA(rng_int_t *n, double *a, double *b, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDBINOMIAL(rng_int_t *n, rng_int_t *m, double *p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL DRANDBINOMIALREFERENCE(rng_int_t *m, double *p, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL DRANDBLUMBLUMSHUB(rng_int_t *n, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDCAUCHY(rng_int_t *n, double *a, double *b, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDCHISQUARED(rng_int_t *n, rng_int_t *df, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDDISCRETEUNIFORM(rng_int_t *n, rng_int_t *a, rng_int_t *b, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL DRANDEXPONENTIAL(rng_int_t *n, double *a, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDF(rng_int_t *n, rng_int_t *df1, rng_int_t *df2, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDGAMMA(rng_int_t *n, double *a, double *b, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDGAUSSIAN(rng_int_t *n, double *xmu, double *var, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDGENERALDISCRETE(rng_int_t *n, double *ref, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL DRANDGEOMETRIC(rng_int_t *n, double *p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL DRANDGEOMETRICREFERENCE(double *p, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL DRANDHYPERGEOMETRIC(rng_int_t *n, rng_int_t *np, rng_int_t *ns, rng_int_t *m, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL DRANDHYPERGEOMETRICREFERENCE(rng_int_t *np, rng_int_t *ns, rng_int_t *m, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL DRANDINITIALIZE(rng_int_t *genid, rng_int_t *subid, rng_int_t *seed, rng_int_t *lseed, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void RNGSTDCALL DRANDINITIALIZEBBS(rng_int_t *nbits, rng_int_t *lp, rng_int_t *p, rng_int_t *lq, rng_int_t *q, rng_int_t *ls, rng_int_t *s, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void RNGSTDCALL DRANDINITIALIZEUSER(RNG_DRANDINITIALIZEUSER_UINI uini, RNG_DRANDINITIALIZEUSER_UGEN ugen, rng_int_t genid, rng_int_t subid, rng_int_t *seed, rng_int_t *lseed, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void RNGSTDCALL DRANDLEAPFROG(rng_int_t *n, rng_int_t *k, rng_int_t *state, rng_int_t *info);
void RNGSTDCALL DRANDLOGISTIC(rng_int_t *n, double *a, double *b, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDLOGNORMAL(rng_int_t *n, double *xmu, double *var, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDMULTINOMIAL(rng_int_t *n, rng_int_t *m, double *p, rng_int_t *k, rng_int_t *state, rng_int_t *x, rng_int_t *ldx, rng_int_t *info);
void RNGSTDCALL DRANDMULTINORMALREFERENCE(rng_int_t *m, double *xmu, double *c, rng_int_t *ldc, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL DRANDMULTINORMALR(rng_int_t *n, double *ref, rng_int_t *state, double *x, rng_int_t *ldx, rng_int_t *info);
void RNGSTDCALL DRANDMULTISTUDENTSTREFERENCE(rng_int_t *m, rng_int_t *df, double *xmu, double *c, rng_int_t *ldc, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL DRANDMULTISTUDENTSTR(rng_int_t *n, double *ref, rng_int_t *state, double *x, rng_int_t *ldx, rng_int_t *info);
void RNGSTDCALL DRANDNEGATIVEBINOMIAL(rng_int_t *n, rng_int_t *m, double *p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL DRANDNEGATIVEBINOMIALREFERENCE(rng_int_t *m, double *p, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL DRANDPOISSON(rng_int_t *n, double *lambda, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL DRANDPOISSONREFERENCE(double *lambda, double *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL DRANDSKIPAHEAD(rng_int_t *n, rng_int_t *state, rng_int_t *info);
void RNGSTDCALL DRANDSTUDENTST(rng_int_t *n, rng_int_t *df, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDTRIANGULAR(rng_int_t *n, double *xmin, double *xmed, double *xmax, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDUNIFORM(rng_int_t *n, double *a, double *b, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDVONMISES(rng_int_t *n, double *vk, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDWEIBULL(rng_int_t *n, double *a, double *b, rng_int_t *state, double *x, rng_int_t *info);
void RNGSTDCALL DRANDMULTINORMAL(rng_int_t *n, rng_int_t *m, double *xmu, double *c, rng_int_t *ldc, rng_int_t *state, double *x, rng_int_t *ldx, rng_int_t *info);
void RNGSTDCALL DRANDMULTISTUDENTST(rng_int_t *n, rng_int_t *m, rng_int_t *df, double *xmu, double *c, rng_int_t *ldc, rng_int_t *state, double *x, rng_int_t *ldx, rng_int_t *info);
void RNGSTDCALL SRANDBETA(rng_int_t *n, float *a, float *b, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDBLUMBLUMSHUB(rng_int_t *n, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDBINOMIAL(rng_int_t *n, rng_int_t *m, float *p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL SRANDBINOMIALREFERENCE(rng_int_t *m, float *p, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL SRANDCAUCHY(rng_int_t *n, float *a, float *b, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDCHISQUARED(rng_int_t *n, rng_int_t *df, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDDISCRETEUNIFORM(rng_int_t *n, rng_int_t *a, rng_int_t *b, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL SRANDEXPONENTIAL(rng_int_t *n, float *a, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDF(rng_int_t *n, rng_int_t *df1, rng_int_t *df2, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDGAMMA(rng_int_t *n, float *a, float *b, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDGAUSSIAN(rng_int_t *n, float *xmu, float *var, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDGENERALDISCRETE(rng_int_t *n, float *ref, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL SRANDGEOMETRIC(rng_int_t *n, float *p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL SRANDGEOMETRICREFERENCE(float *p, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL SRANDHYPERGEOMETRIC(rng_int_t *n, rng_int_t *np, rng_int_t *ns, rng_int_t *m, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL SRANDHYPERGEOMETRICREFERENCE(rng_int_t *np, rng_int_t *ns, rng_int_t *m, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL SRANDINITIALIZE(rng_int_t *genid, rng_int_t *subid, rng_int_t *seed, rng_int_t *lseed, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void RNGSTDCALL SRANDINITIALIZEBBS(rng_int_t *nbits, rng_int_t *lp, rng_int_t *p, rng_int_t *lq, rng_int_t *q, rng_int_t *ls, rng_int_t *s, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void RNGSTDCALL SRANDINITIALIZEUSER(RNG_SRANDINITIALIZEUSER_UINI uini, RNG_SRANDINITIALIZEUSER_UGEN ugen, rng_int_t genid, rng_int_t subid, rng_int_t *seed, rng_int_t *lseed, rng_int_t *state, rng_int_t *lstate, rng_int_t *info);
void RNGSTDCALL SRANDLEAPFROG(rng_int_t *n, rng_int_t *k, rng_int_t *state, rng_int_t *info);
void RNGSTDCALL SRANDLOGISTIC(rng_int_t *n, float *a, float *b, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDLOGNORMAL(rng_int_t *n, float *xmu, float *var, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDMULTINOMIAL(rng_int_t *n, rng_int_t *m, float *p, rng_int_t *k, rng_int_t *state, rng_int_t *x, rng_int_t *ldx, rng_int_t *info);
void RNGSTDCALL SRANDMULTINORMALREFERENCE(rng_int_t *m, float *xmu, float *c, rng_int_t *ldc, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL SRANDMULTINORMALR(rng_int_t *n, float *ref, rng_int_t *state, float *x, rng_int_t *ldx, rng_int_t *info);
void RNGSTDCALL SRANDMULTISTUDENTSTREFERENCE(rng_int_t *m, rng_int_t *df, float *xmu, float *c, rng_int_t *ldc, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL SRANDMULTISTUDENTSTR(rng_int_t *n, float *ref, rng_int_t *state, float *x, rng_int_t *ldx, rng_int_t *info);
void RNGSTDCALL SRANDNEGATIVEBINOMIAL(rng_int_t *n, rng_int_t *m, float *p, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL SRANDNEGATIVEBINOMIALREFERENCE(rng_int_t *m, float *p, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL SRANDPOISSON(rng_int_t *n, float *lambda, rng_int_t *state, rng_int_t *x, rng_int_t *info);
void RNGSTDCALL SRANDPOISSONREFERENCE(float *lambda, float *ref, rng_int_t *lref, rng_int_t *info);
void RNGSTDCALL SRANDSKIPAHEAD(rng_int_t *n, rng_int_t *state, rng_int_t *info);
void RNGSTDCALL SRANDSTUDENTST(rng_int_t *n, rng_int_t *df, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDTRIANGULAR(rng_int_t *n, float *xmin, float *xmed, float *xmax, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDUNIFORM(rng_int_t *n, float *a, float *b, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDVONMISES(rng_int_t *n, float *vk, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDWEIBULL(rng_int_t *n, float *a, float *b, rng_int_t *state, float *x, rng_int_t *info);
void RNGSTDCALL SRANDMULTINORMAL(rng_int_t *n, rng_int_t *m, float *xmu, float *c, rng_int_t *ldc, rng_int_t *state, float *x, rng_int_t *ldx, rng_int_t *info);
void RNGSTDCALL SRANDMULTISTUDENTST(rng_int_t *n, rng_int_t *m, rng_int_t *df, float *xmu, float *c, rng_int_t *ldc, rng_int_t *state, float *x, rng_int_t *ldx, rng_int_t *info);

/* RNG version information */
const char* RNGSTDCALL GET_RNGVERSION(void);


#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif  /* _RNG_H */
